'''
Created on Feb 3, 2016

@author: mukuze
'''

'''imports'''

# other built in modules
import numpy as np
from copy import deepcopy
import operator
import sys

# import from other modules
import Exact_bool_factorization as ebf
import Merging as mrg


# global variables
input_mat=[]
indx=[]
nb_cols=0
nb_rows=0
# constraints=[]

''' Function that accepts values from the main GUI_interactiveBMF frame (class) and sets the global 
variables to be used by the functions in this module.
Only the indices of 1s and input matrix need to be set accordingly so the results returned by some functions in this module are right'''
def get_set_values(full_mat, covered_indices, number_cols, number_rows):
    global input_mat
    input_mat=full_mat
    
    global indx
    indx=covered_indices
    
    global nb_cols
    nb_cols=number_cols
    
    global nb_rows
    nb_rows=number_rows
    
    # here the system is basically run without constraints that is why the matrix is a matrix of zeros
    # to run the system with constraints a matrix with 1s in the entries we want to keep as they are (no error allowed)
    # and 0s in the entries whose values we don't care about (error is allowed) should be given as input
    # to run the system with constraints, uncomment line below
#     global constraints
#     constraints=np.zeros_like(input_mat)
    
''' Function that sorts the factor_items in increasing number of 1's they cover alone. 
This results in the factor that covers the least ones comes first in the dictionary'''     
def sort_number_1s(scores):
    sorted_scores = sorted(scores.items(), key=operator.itemgetter(1))
    return sorted_scores
 

''' Function that returns the matrix multiplication of the two rank-1 matrices corresponding to the indexed factor_item. 
The first rank-1 matrix is the boolean matrix with 1 at the indexes covered by the closed itemset itself and
the second is the boolean matrix with 1 at the indexes in the transactions containing the closed itemset'''
def vec_gen(clitems,trans_sets,factoritem, indx,nb_cols,nb_rows):
    mylist1=list(clitems[factoritem[indx]])
    mylist2=list(trans_sets[factoritem[indx]])
    row_vec=np.zeros((1,nb_cols))
    col_vec=np.zeros((nb_rows,1))
    for i in mylist1:
        row_vec[0][i]=1
    for j in mylist2:
        col_vec[j]=1
    return col_vec.dot(row_vec)
    
''' Function that generates the full data by doing an OR operation between the matrices generated by the rank-1 vectors corresponding to the closed item and its transactions.
In this case every entry that is covered by at least one of the rank-1 matrices is assigned 1 and any entry is not covered by any of the rank-1 matrices then it is assigned 0''' 
def gen_matrix(clitems,trans_sets,factoritem,nb_cols,nb_rows):
    mat=vec_gen(clitems,trans_sets,factoritem,0,nb_cols,nb_rows)
    full_mat=np.zeros_like(mat)
    for x in range(len(factoritem)):
        full_mat+= vec_gen(clitems,trans_sets,factoritem,x,nb_cols,nb_rows)
    full_mat[full_mat>1]=1
    return full_mat


''' Function that finds the entries where the covering makes errors (and set them to 1)
Inputs: the original matrix, and the matrix generated from adding the rank-1 matrices resulting from the factor items
This function is called after every merge or delete operation'''
def error_mat(input_mat,generated_mat):
    err_mat=np.absolute(np.subtract(input_mat,generated_mat))
    return err_mat

''' Function that finds if we make an error on any of the entries which is part of the constraints.
The constraints matrix has entry to be 1 if we want the entry to stay as it is (no error allowed), and 0 if we don't care less about the value (we allow error).
Inputs: the error matrix which has 1 if we make an error on that entry, the constraints matrix
Output: true if we make an error and false if we don't make an error on a constrained entry'''
def made_error(err_mat,constraints):
    # multiply both matrices, the result is 0 if no constraint is violated and 1 if a constraint is violated
    new_mat=np.multiply(err_mat,constraints)
    if np.any(new_mat==1):
        return True
    else:
        return False

''' Function that generated the UNION of the two factor_items with highest similarity.
The function takes a similar pair, the previous trans_items, closed_items and factor_items lists as input. 
The output is a modified trans_items, closed_items, and factor_items list since we have merged some factor_items'''
def merge_factor(similar_pair,clitems,trans_sets,factoritems):
    
    f_copy = deepcopy(factoritems)
    t_copy=deepcopy(trans_sets)
    cl_copy=deepcopy(clitems)
    union_transitems=t_copy[f_copy[similar_pair[0]]].union(t_copy[f_copy[similar_pair[1]]])
    union_clitems=cl_copy[f_copy[similar_pair[0]]].union(cl_copy[f_copy[similar_pair[1]]])
    f_copy[similar_pair[0]]=-1
    f_copy[similar_pair[1]]=-1
    t_copy.append(union_transitems)
    cl_copy.append(union_clitems)   
        
    f_copy=[x for x in f_copy if x!=-1]
    last_index=len(cl_copy)-1
    f_copy.append(last_index)
     
    return t_copy,cl_copy,f_copy 

''' Function that generated the INTERSECTION of the two factor_items with highest similarity.
The function takes a similar pair, the previous trans_items, closed_items and factor_items lists as input. 
The output is a modified trans_items, closed_items, and factor_items list since we have merged some factor_items'''
def merge_factor_inter(similar_pair,clitems,trans_sets,factoritems):
    
    f_copy = deepcopy(factoritems)
    t_copy=deepcopy(trans_sets)
    cl_copy=deepcopy(clitems)
    
    inter_transitems=t_copy[f_copy[similar_pair[0]]].intersection(t_copy[f_copy[similar_pair[1]]])
    inter_clitems=cl_copy[f_copy[similar_pair[0]]].intersection(cl_copy[f_copy[similar_pair[1]]])
    f_copy[similar_pair[0]]=-1
    f_copy[similar_pair[1]]=-1
    t_copy.append(inter_transitems)
    cl_copy.append(inter_clitems) 
        
    f_copy=[x for x in f_copy if x!=-1]
    last_index=len(cl_copy)-1
    f_copy.append(last_index)
     
    return t_copy,cl_copy,f_copy 
 
''' Function that computes the error that we make by deleting a factor item'''
def error_del(delete_indx, scores, factoritems):
    return scores[factoritems[delete_indx]]

''' Function that computes the merge of a specific pair and returns the errors as well as the resulting closed items, trans sets and factors for both OR and AND operations (optimized)'''
def error_OR_AND_optimized(merge_pair,clitems, trans_sets, factoritems):
    global indx
    
    f1_copy = deepcopy(factoritems)
    t1_copy=deepcopy(trans_sets)
    cl1_copy=deepcopy(clitems)
    f2_copy = deepcopy(factoritems)
    t2_copy=deepcopy(trans_sets)
    cl2_copy=deepcopy(clitems)
    
    # indices of 1s in data
    orig_indx=indx
    
    # indices of the merged factor with OR
    new_trans=trans_sets[factoritems[merge_pair[0]]].union(trans_sets[factoritems[merge_pair[1]]])
    new_items=clitems[factoritems[merge_pair[0]]].union(clitems[factoritems[merge_pair[1]]])
    new_factor_or= set((x,y) for x in new_trans for y in new_items)
    
    # error for OR
    errors_OR=len(new_factor_or.difference(map(tuple,orig_indx)))
    
    # indices of the merged factor with AND
    new_trans_and=trans_sets[factoritems[merge_pair[0]]].intersection(trans_sets[factoritems[merge_pair[1]]])
    new_items_and=clitems[factoritems[merge_pair[0]]].intersection(clitems[factoritems[merge_pair[1]]])
    new_factor_and= set((x,y) for x in new_trans_and for y in new_items_and)
    
    # indices of potential 1s turning into 0s due to merge with AND
    potentials_errors_and=new_factor_or.difference(new_factor_and)
    
    # error for AND
    errors_AND=len(potentials_errors_and.intersection(map(tuple,orig_indx)))
    
    # updates for OR
    # update the transactions and closed items after merging
    t1_copy.append(new_trans)
    cl1_copy.append(new_items)  
    
    # update the factors
    merge_pair_factors=[f1_copy[merge_pair[0]], f1_copy[merge_pair[1]]]
    factors_or=[x for x in f1_copy if x not in merge_pair_factors]
    factors_or.append(len(cl1_copy)-1)
    
    # updates for AND
    t2_copy.append(new_trans_and)
    cl2_copy.append(new_items_and)  
    
    # update the factors
    factors_and=[x for x in f2_copy if x not in merge_pair_factors]
    factors_and.append(len(cl2_copy)-1)
    
    
    return errors_OR, errors_AND, cl1_copy, t1_copy, factors_or, cl2_copy, t2_copy, factors_and
 
''' Function that computes the merge of a specific pair and returns the errors as well as the resulting closed items, trans sets and factors for both OR and AND operations'''
def error_OR_AND(merge_pair,clitems, trans_sets, factoritems): 
    global indx
    
    orig_indx=indx
    # merge with OR operator
    # check the error by checking how many 0's turned into 1's
    trans,closed,factors=merge_factor(merge_pair,clitems,trans_sets,factoritems)
    
    # full matrix with merge with OR
    # generate full matrix by adding the 1's covered by the factors
    full_mat_or=gen_matrix(closed,trans,factors,nb_cols,nb_rows)
    
    # indices of 1's in current full data
    current_indx_=np.transpose(np.nonzero(full_mat_or))
    
    indices_itemsets_1=[(x,y) for x in trans_sets[factoritems[merge_pair[0]]] for y in clitems[factoritems[merge_pair[0]]]]
    indices_itemsets_2=[(x,y) for x in trans_sets[factoritems[merge_pair[1]]] for y in clitems[factoritems[merge_pair[1]]]]
    
    # indices covered by both itemsets to be merged
    indices_itemsets=set(map(tuple,indices_itemsets_1)).union(map(tuple,indices_itemsets_2))
    
    # indices covered by the current itemset or factor after OR operation
    indices_itemset_or=[(x,y) for x in trans[factors[-1]] for y in closed[factors[-1]]]
    
    # indices which changed to 1 after OR operation
    indices_diff=set(map(tuple,indices_itemset_or)).difference(indices_itemsets)
    # indices which are originally zero in original data
    indices_orig_zeros=indices_diff.difference(map(tuple,orig_indx))
    # which of the original zeros are now 1s in current data
    indices_zeros_ones=indices_orig_zeros.intersection(map(tuple,current_indx_))
    
    indices_zeros_ones=map(list,indices_zeros_ones)
    
    error_or=len(indices_zeros_ones)
    
    # merge with AND operator
    # check the error by checking how many 1's turned into 0's
    # we check the indices originally on the rows where the factor item is present
    trans_and,closed_and,factors_and=merge_factor_inter(merge_pair,clitems,trans_sets,factoritems)
    # full matrix with merge with AND
    # generate full matrix by adding the 1's covered by the factors
    full_mat_and=gen_matrix(closed_and,trans_and,factors_and,nb_cols,nb_rows)
    
    # indices of 1's in current full data
    current_indx_and=np.transpose(np.nonzero(full_mat_and))
    
    # indices covered by the current itemset or factor after OR operation
    indices_itemset_and=[(x,y) for x in trans_and[factors_and[-1]] for y in closed_and[factors_and[-1]]]
    
    # indices which changef to 0 after AND operation
    indices_diff_and=set(map(tuple,indices_itemsets)).difference(map(tuple,indices_itemset_and))
    
    # indices which are originally ones in original data
    indices_orig_ones=indices_diff_and.intersection(map(tuple,orig_indx))
    
    
    indices_ones_zeros=indices_orig_ones.difference(map(tuple,current_indx_and))

    indices_missing=set(map(tuple,current_indx_and)).difference(map(tuple,orig_indx))
    indices_corrected=indices_missing.intersection(indices_ones_zeros)
    indices_corrected=map(list,indices_corrected)
    
    indices_ones_zeros=map(list,indices_ones_zeros)
                     
    actual_error_and=len(indices_ones_zeros)-len(indices_corrected)
    
    # return the errors and as well the closed items,  trans_sets, and factors if the recommended pair were to be merged
    # this will help in showing the user what would be happen should be go ahead with the merge
    return error_or,actual_error_and, closed, trans, factors, closed_and, trans_and, factors_and

''' Function that computes the delete for a specific index and merge for a specific pair of factors and returns the best operation that 
does not violate the constraints in case the constraints are provided as input. When the constraints are not provided as input, the
function still returns the best operation'''
def best_action(delete_indx, merge_pair, clitems, trans_sets, factoritems, scores):
    # local variables
#     global acc_error
    
    # uncomment this part of the code if you are giving the system a system of constraints as input
    # so then this piece of code checks if deleting the factor violates any user specified constraints.
    
#     # remove the factor to delete
#     factors=[x for x in factoritems if x!=factoritems[delete_indx]]   
    
#     # compute the full matrix after removing the factor to delete
#     gen_mat=gen_matrix(clitems,trans_sets,factors,nb_cols,nb_rows)
#     
#     # check if that does not violate any constraint
#     err_mat_del=error_mat(input_mat,gen_mat)
#     
#     del_chk=made_error(err_mat_del,constraints)
    
    # error of the delete operation
    error_delete=error_del(delete_indx, scores, factoritems)
           
    if merge_pair!="":
    
        # merge the factors 
        # merge with OR operator
        # check the error by checking how many 0's turned into 1's
#         trans_or,closed_or,factors_or=merge_factor(merge_pair,clitems,trans_sets,factoritems)
        
        # uncomment this part of the code if you are giving the system a system of constraints as input
        # so then this piece of code checks if deleting the factor violates any user specified constraints.
        
#         # full matrix with merge with OR
#         # generate full matrix by adding the 1's covered by the factors
#         full_mat_or=gen_matrix(closed_or,trans_or,factors_or,nb_cols,nb_rows)
#         
#         # check if that does not violate any constraint
#         err_mat_or=error_mat(input_mat,full_mat_or) 
#             
#         merg_or_chk=made_error(err_mat_or,constraints)
                
        # merge with AND operator
        # check the error by checking how many 1's turned into 0's
        # we check the indices originally on the rows where the factor item is present
#         trans_and,closed_and,factors_and=merge_factor_inter(merge_pair,clitems,trans_sets,factoritems)
        
        # uncomment this part of the code if you are giving the system a system of constraints as input
        # so then this piece of code checks if deleting the factor violates any user specified constraints.
        
        # full matrix with merge with AND
        # generate full matrix by adding the 1's covered by the factors
#         full_mat_and=gen_matrix(closed_and,trans_and,factors_and,nb_cols,nb_rows)
#         
#         # check if that does not violate any constraint
#         err_mat_and=error_mat(input_mat,full_mat_and)
#        
#         merg_and_chk= made_error(err_mat_and,constraints)
        
#         # we are going to return all the errors so we need not compute them again in case the user skips the currently recommended operation
#         # if the recommended operation is a Merge with AND, we return the errors as following; error_merge_and, ..., error_merge_or, error_delete
#         # if the recommended operation is a Merge with OR, we return the errors as following; error_merge_or, ..., error_merge_and, error_delete
#         # if the recommended operation is a Delete, we return the errors as following; error_delete, ..., error_merge_or, error_merge_and  
#         if del_chk:
#             if merg_or_chk:
#                 if merg_and_chk:
#                     return "", "", "", "", "", "", "", ""
#                 else:
# #                     acc_error+=error_merge_and
#                     return "a", error_merge_and, merge_pair, closed_and, trans_and, factors_and, error_merge_or, error_delete 
#             else:
#                 if merg_and_chk:
# #                     acc_error+=error_merge_or
#                     return "o", error_merge_or, merge_pair, closed_or, trans_or, factors_or, error_merge_and, error_delete
#                 else:
#                     if error_merge_or<=error_merge_and:
# #                         acc_error+=error_merge_or
#                         return "o", error_merge_or, merge_pair, closed_or, trans_or, factors_or, error_merge_and, error_delete
#                     else:
# #                         acc_error+=error_merge_and
#                         return "a", error_merge_and, merge_pair, closed_and, trans_and, factors_and, error_merge_or, error_delete 
#         else:
#             if merg_or_chk:
#                 if merg_and_chk:
# #                     acc_error+=error_delete
#                     return "d", error_delete, delete_indx, "", "", "", error_merge_or, error_merge_and
#                 else:
#                     if error_delete<=error_merge_and:
# #                         acc_error+=error_delete
#                         return "d", error_delete, delete_indx, "", "", "", error_merge_or, error_merge_and
#                     else:
# #                         acc_error+=error_merge_and
#                         return "a", error_merge_and, merge_pair, closed_and, trans_and, factors_and, error_merge_or, error_delete   
#             else:
#                 if merg_and_chk:
#                     if error_delete<=error_merge_or:
# #                         acc_error+=error_delete
#                         return "d", error_delete, delete_indx, "", "", "", error_merge_or, error_merge_and
#                     else:
# #                         acc_error+=error_merge_or
#                         return "o", error_merge_or, merge_pair, closed_or, trans_or, factors_or, error_merge_and, error_delete
#                 else:
#                     if error_delete<=error_merge_or and error_delete<=error_merge_and:
# #                         acc_error+=error_delete
#                         return "d", error_delete, delete_indx, "", "", "",  error_merge_or, error_merge_and
#                     
#                     elif error_merge_or<=error_delete and error_merge_or<=error_merge_and:
# #                         acc_error+=error_merge_or
#                         return "o", error_merge_or, merge_pair, closed_or, trans_or, factors_or, error_merge_and, error_delete
#                     
#                     elif error_merge_and<=error_delete and error_merge_and<=error_merge_or:
# #                         acc_error+=error_merge_and
#                         return "a", error_merge_and, merge_pair, closed_and, trans_and, factors_and, error_merge_or, error_delete  
#     else:
#         error_delete=error_del(delete_indx, scores, factoritems)
#         if del_chk:
#             return "", "", "",  "", "", "", "", ""
#         
#         else:
# #             acc_error+=error_delete
#             return "d", error_delete, delete_indx,  "", "", "", sys.maxint, sys.maxint
 
        # comment the few lines of code below if the user predefines the constraints and uncomment the code above
        # error of the merge with OR and AND operations
        error_merge_or, error_merge_and, closed_or, trans_or, factors_or, closed_and, trans_and, factors_and=error_OR_AND(merge_pair,clitems, trans_sets, factoritems)
        
        if error_delete<=error_merge_or:
            if  error_delete<=error_merge_and:
                return "d", error_delete, delete_indx, "", "", "",  error_merge_or, error_merge_and
            
            else:
                return "a", error_merge_and, merge_pair, closed_and, trans_and, factors_and, error_merge_or, error_delete
        
        else:
            if error_merge_or<=error_merge_and:
                return "o", error_merge_or, merge_pair, closed_or, trans_or, factors_or, error_merge_and, error_delete
            
            else:
                return "a", error_merge_and, merge_pair, closed_and, trans_and, factors_and, error_merge_or, error_delete
         
    else:
        return "d", error_delete, delete_indx,  "", "", "", sys.maxint, sys.maxint
        
''' Function that returns the updated scores and number of 1s '''
def updates_OR_AND(clitems,trans_sets,old_factors,new_factors,scores,number_1s,reducedindxcover,operation,pair_to_merge):
    
    # make a copy of the scores and the number of 1s
    scores_copy=deepcopy(scores)
    number_1s_copy=deepcopy(number_1s)
    reducedindxcover_copy=deepcopy(reducedindxcover)
    
    if operation=="a":
        # we need to update the number of 1s
        # delete first the keys and values corresponding to the pair to merge
        for x in pair_to_merge:
            del number_1s_copy[old_factors[x]]
            
        # update the number of 1s
        number_1s_copy[new_factors[-1]]=len(trans_sets[-1])*len(clitems[-1])
        
        # we need to update the scores
        # first get the indices that each factor in the merge pair covers
        indx_factor_1=[(x,y) for x in trans_sets[old_factors[pair_to_merge[0]]] for y in clitems[old_factors[pair_to_merge[0]]]]
        indx_factor_2=[(x,y) for x in trans_sets[old_factors[pair_to_merge[1]]] for y in clitems[old_factors[pair_to_merge[1]]]]
        
        # the length of the intersection of the two sets of indices above will give us the number of 1s that the merged factor covers
        merge_pair_unique_area=len(set(indx_factor_1).intersection(set(indx_factor_2)))
        
        # delete first the keys and values corresponding to the pair to merge
        for x in pair_to_merge:
            del scores_copy[old_factors[x]]
            
        # factors in merge pair
        merge_pair_factors=[old_factors[x] for x in pair_to_merge]
        
        # update the scores
        # by deleting some factors, the number of ones covered alone by some other factors may increase
        # so we look through the reduced_cover and update the scores accordingly
        for k in reducedindxcover_copy.keys():
            # we only consider the indices which are covered only by 2 factors since these will be covered by only 1 if the other is deleted
            if len(reducedindxcover_copy[k])==2:
                  
                factor_to_update=list(set(reducedindxcover_copy[k]).difference(set(merge_pair_factors)))
                  
                # if the difference is not empty then it contains the factor whose score should be incremented by 1
                if len(factor_to_update)==1:
                    if factor_to_update[0] in scores_copy.keys():
                        
                        scores_copy[factor_to_update[0]]+=1
                        
        
        # then update the scores
        scores_copy[new_factors[-1]]=merge_pair_unique_area
        
        return scores_copy, number_1s_copy
    
    else:
        # we need to update the number of 1s
        # delete first the keys and values corresponding to the pair to merge
        for x in pair_to_merge:
            del number_1s_copy[old_factors[x]]
       
       
        number_1s_copy[new_factors[-1]]=len(trans_sets[-1])*len(clitems[-1])
        
        # we need to update the scores
        # first get the number of 1s that each factor in the pair to merge covers alone
        merge_pair_unique_area=0
    
        for x in pair_to_merge:
            merge_pair_unique_area+=scores_copy[old_factors[x]]
            
        # delete first the keys and values corresponding to the pair to merge
        for x in pair_to_merge:
            del scores_copy[old_factors[x]]
            
        # then update the scores
        # if the new factor covers no one alone there is no need to update the scores
       
        scores_copy[new_factors[-1]]=merge_pair_unique_area
        
        return scores_copy, number_1s_copy
        
    
            
''' Function that updates the closed items, transactions,  scores, cover, and factors in the case the user says yes to go ahead with the recommended operation'''
def update(clitems,trans_sets,factoritems,scores,cover,reducedindxcover,operation,remove_indx,pair_to_merge):
    # update the matrices according to the returned operation
    if operation=="d":
        # We proceed with delete operation
        # remove the factor which covers the less number of 1s
        factors=[x for x in factoritems if x!=factoritems[remove_indx]] 
        # modify the reduced_cover to reflect removal
        for k in reducedindxcover.keys():
            if factoritems[remove_indx] in reducedindxcover[k]:
                reducedindxcover[k].remove(factoritems[remove_indx]) 
        
        # update the scores
        scores=ebf.score(reducedindxcover, factors)
        # update factors
        factoritems=factors
        
        # we do not return the transaction sets (trans_sets) and closed items (clitems) cause the delete operation does not change them
        return factoritems, reducedindxcover, scores
        
    else:
        
        # The closed items, trans_sets, and factor items are already updated when calling the best_action_perform function
        # So we only update here the scores and reduced_index_cover

        # generate new scores with updated factor_items
        # update the indices_cover dictionary to reflect the changes in transactions and closed items. The update makes us avoid recomputing the whole dictionary all over
        indices_new_item_set=[(x,y) for x in trans_sets[-1] for y in clitems[-1]]
        
        for x in indices_new_item_set:
            if x in cover.keys():
                cover[x].add(factoritems[-1])
            else:
                cover[x]=set()
                cover[x].add(factoritems[-1])
            
        reducedindxcover=ebf.reduced_cover(cover, factoritems)
        scores=ebf.score(reducedindxcover, factoritems)
        scores={k:v for k,v in scores.iteritems() if scores[k]>0}
        factoritems=[x for x in scores.keys()]
        
        # here we return all the arguments cause the merge operation changes the contents of all the arguments
        return clitems, trans_sets, factoritems, reducedindxcover, scores
 
''' Function that performs the action returned above'''
def best_action_perform(clitems,trans_sets,factoritems,scores):
    # global variable
    global t_best_move
    # local variable
    sim_pairs={}
    returned_action= "", "", "",  "", "", ""
    
    if len(factoritems)>1:
#     if len(factoritems)!=89:
        # select the index to delete
        # sort the scores to see which factor item covers less number of 1s alone
        desc_scores=sort_number_1s(scores)
        
        # index to delete
        indx_del=factoritems.index(desc_scores[0][0])
        
        # return closed pairs and choose the pair to be merged
        # proceed with merges
        mrg.get_set_values(clitems, trans_sets, factoritems, nb_rows, nb_cols)
        sim_pairs=mrg.chosen_sim_pairs(mrg.minHash)
        
#     else:
# #         ranks.append(len(factoritems))
# #         errors.append(acc_error)
#         print ""
# #         print "Only one factor left!"
#         print "The number of 1s in the data is:"
#         print len(indx)
#         print "The number of factors is:"
#         print len(factoritems)
#         print "The error is:"
#         print acc_error
    
    if sim_pairs:
        # convert dictionary keys into tuples 
        similar_p=tuple(sim_pairs.keys())
        # pair to merge
        pair_merge=similar_p[0]
        
        # send the index to delete and pair to merge to the function above to return the best option
#         start_time = datetime.now() 
        
        returned_action=best_action(indx_del, pair_merge, clitems, trans_sets, factoritems, scores)

#         end_time = datetime.now()

#         t_span=end_time-start_time
#         t_best_move.append(t_span)
        
    else:
        # send the index to delete and pair to merge to the function above to return the best option
#         start_time = datetime.now() 
        
        returned_action=best_action(indx_del, "", clitems, trans_sets, factoritems, scores)
        
#         end_time = datetime.now()
        
#         t_span=end_time-start_time
#         t_best_move.append(t_span)
        
    if returned_action[0]=="":
#         ranks.append(len(factoritems))
#         errors.append(acc_error)
        
        # we return the similar pairs and scores in increasing order so we do not need to recompute them later when the user decides not to merge the first suggested pair
        return "", "", "",  "", "", "", "", "", similar_p, desc_scores
        
    elif returned_action[0]=="d":
        
        # return the recommended operation returned_action[0],
        # the error of the operation returned_action[1],
        # the index to delete returned_action[2],
        # the closed items returned_action[3],
        # the trans_sets returned_action[4],
        # the factor items returned_action[5],
        # we also return the errors of all the other operations so that if the user decides not to commit the recommended best operation we do not need
        # to recompute these errors once again in order to recommend the next best operation
        # the error of the merge with OR for the first similar pair returned_action[6]
        # the error of the merge with AND for the first similar pair returned_action[7]
        # the accumulated error
        # and the similar pairs and scores in increasing order so we do not need to recompute them later when the user decides not to merge the first suggested pair
        # check if the similar pairs variable is not empty, if it is empty save an empty tuple of tuples
        if sim_pairs:
            return returned_action[0], returned_action[1], returned_action[2], returned_action[3],returned_action[4],returned_action[5], returned_action[6], returned_action[7], similar_p, desc_scores
        else:
            return returned_action[0], returned_action[1], returned_action[2], returned_action[3],returned_action[4],returned_action[5], returned_action[6], returned_action[7], (), desc_scores
        
    
        
    elif returned_action[0]=="a":
        
        # return the recommended operation returned_action[0],
        # the error of the operation returned_action[1],
        # the index to delete returned_action[2],
        # the closed items returned_action[3],
        # the trans_sets returned_action[4],
        # the factor items returned_action[5],
        # we also return the errors of all the other operations so that if the user decides not to commit the recommended best operation we do not need
        # to recompute these errors once again in order to recommend the next best operation
        # the error of the merge with OR for the first similar pair returned_action[6]
        # the error of the Delete operation returned_action[7]
        # the accumulated error
        # and the similar pairs and scores in increasing order so we do not need to recompute them later when the user decides not to merge the first suggested pair
        return returned_action[0], returned_action[1], returned_action[2], returned_action[3],returned_action[4],returned_action[5], returned_action[6], returned_action[7], similar_p, desc_scores
        
        
    else:
        
        # return the recommended operation returned_action[0],
        # the error of the operation returned_action[1],
        # the index to delete returned_action[2],
        # the closed items returned_action[3],
        # the trans_sets returned_action[4],
        # the factor items returned_action[5],
        # we also return the errors of all the other operations so that if the user decides not to commit the recommended best operation we do not need
        # to recompute these errors once again in order to recommend the next best operation
        # the error of the merge with AND for the first similar pair returned_action[6]
        # the error of the Delete operation returned_action[7]
        # the accumulated error
        # and the similar pairs and scores in increasing order so we do not need to recompute them later when the user decides not to merge the first suggested pair
        return returned_action[0], returned_action[1], returned_action[2], returned_action[3],returned_action[4],returned_action[5], returned_action[6], returned_action[7], similar_p, desc_scores

# print ""
# print "Interactive part starts"
# results=best_action_perform(cl_items,trans_items,factor_items,scores,reduced_indx_cover)
# print results[0],results[1],results[2],results[3]