#! /usr/bin/env python
'''
Created on Nov 30, 2015

@author: mukuze

Description: A script to preprocess the data in order to put it in the right format
for the fpgrowth algorithm used to generate the frequent closed itemsets.
After meeting with Pauli on 13 April, 2016, we decided to extend the script 
so that it also generates the closed itemsets by running the fp-growth algorithm,
and also does the exact factorization and returns the various files needed to proceed
with the interactive part of the program. 
'''

import pandas as pd
import numpy as np
import sys
from argparse import ArgumentParser
import argparse
import os.path
from datetime import datetime
import cPickle
import os

# import needed to be able to run a shell script in a python script
from subprocess import Popen, call, PIPE

# import extra module
import Exact_bool_factorization as ebf

def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file %s does not exist!" % arg)
    else:
        return open(arg, 'r')  # return an open file handle
    

def check_value(value):
    ivalue = int(value)
    if ivalue not in [0,1]:
        raise argparse.ArgumentTypeError("%s is an invalid. Should be 0 or 1" % value)
    return ivalue

def check_minsup(parser, sup):
    isup = int(sup)
    if isup>100 or isup<1:
        raise argparse.ArgumentTypeError("%s is invalid. Minimum support should be between 1 and 100" % sup)
    return isup

# parser = ArgumentParser(description="Script to preprocess the input data.\n"+
#                         "The input data should be boolean and preferably with columns and rows\n"+
#                         "named. The columns should be IN ANY CASE named.")

parser = ArgumentParser(description="Script to preprocess the input data.\n"+
                        "The input data should be binary (0/1) and in csv (comma separated value) format")

# parser.add_argument("-i", dest="filename", required=True,
#                     help="Input file in the right format (both column and rows have names or only column have names", metavar="FILE",
#                     type=lambda x: is_valid_file(parser, x))
requiredNamed = parser.add_argument_group('required named arguments')
requiredNamed.add_argument("-f1",dest="filename1", required=True,
                    help="Input Data in the right format (binary format (0/1) and in csv (comma separated value) format [Required]", metavar="FILE1",
                    type=lambda x: is_valid_file(parser, x))

# parser.add_argument("-rc", type=check_value, required="True",dest="opt",
#                     help="Enter 0 if the first column of your data does not contain the row names and 1 if it does.", metavar="0 or 1",)
# args = parser.parse_args()

parser.add_argument("-f2",dest="filename2", required=False,
                    help="Row names in the csv format [optional]", metavar="FILE2",
                    type=lambda y: is_valid_file(parser, y))

requiredNamed.add_argument("-f3",dest="filename3", required=True,
                    help="Column names in the csv format [Required]", metavar="FILE3",
                    type=lambda z: is_valid_file(parser, z))

requiredNamed.add_argument("-sup",dest="minsup", required=True,
                    help="minimum support in percentage of transactions [Required]", metavar="MINIMUM SUPPORT",
                    type=lambda u: check_minsup(parser, u))

args = parser.parse_args()

path_to_location=""

if len(sys.argv)>7:
    try:     
        filename1=sys.argv[2]
        filename2=sys.argv[4]
        filename3=sys.argv[6]
        
        # update the path to the location of the input
        path_to_location=os.path.dirname(filename1)
        
        data=pd.read_csv(filename1,sep=",", header=None)
        data_rows=pd.read_csv(filename2,sep=",", header=None)
        data_cols=pd.read_csv(filename3,sep=",", header=None)
        data=data.values
        headers=[x for x in range(len(data[1,:]))]
        headers=np.asarray(headers)
        headers=headers.reshape(1,len(data[1,:]))
        data_=np.concatenate((headers,data))
#         column_names=list(new_data.columns.values)
        np.savetxt(os.path.dirname(filename1)+"/input",data_,delimiter=" ",fmt="%d")
        np.savetxt(os.path.dirname(filename1)+"/column_names",data_cols.values,delimiter="\n",fmt="%s")
        np.savetxt(os.path.dirname(filename1)+"/row_names",data_rows.values,delimiter="\n",fmt="%s")
        print '\n***Step 1 of Prepocessing successful: The three files "input, column_names, row_names" were written to the root directory. \n'
    except TypeError:
        print "\nType Error: Please provide an input binary file without column names and row names. \n"
        sys.exit(1)
    
    if os.path.exists("fpgrowth/fpgrowth/ex/flg2set")  and os.path.exists("fpgrowth/fpgrowth/ex/fpgrowth"):
    	# Now we try and call the scripts in the src/fp-growth/fp-growth/exec folder in order to mine the closed itemsets
    	# we call first the flg2set script to convert the input of the user to the right format for the fp-growth algorithm
    	print "Now doing some preprocessing before running the fp-growth algorithm to mine all closed itemsets.\n"
    	
    	print "Converting user input to right format for the fp-growth algorithm.\n"
    	
    	print ("Executing: ./fpgrowth/fpgrowth/ex/flg2set " + path_to_location+"/input"+ " " + path_to_location+"/x1.tab")
    	print ""
    	command="./fpgrowth/fpgrowth/ex/flg2set " + path_to_location+"/input"+ " " + path_to_location+"/x1.tab"
    	process=Popen(command, stdout=PIPE, stderr=None, shell=True)
    	#Launch the shell command:
    	process.communicate()
    	
    	print "\n***Step 2 Done converting to the format required by fp-growth.***\n"
    	
    	# we call first the fpgrowth script to generate the closed itemsets using the fp-growth algorithm
    	print "Now mining the closed itemsets using the fp-growth algorithm.\n"
    	
    	print ("Executing: ./fpgrowth/fpgrowth/ex/fpgrowth -tcm2s"+str(sys.argv[8])+"""k"," -v" """""+"\" " + path_to_location+"/x1.tab"+ " " + path_to_location+"/closed_items")
    	print ""
    	FNULL = open(os.devnull, 'w') 
    	command="./fpgrowth/fpgrowth/ex/fpgrowth -tcm2s"+str(sys.argv[8])+"""k"," -v" """""+"\" " + path_to_location+"/x1.tab"+ " " + path_to_location+"/closed_items"
    	process=Popen(command, stdout=PIPE, stderr=None, shell=True)
    	#Launch the shell command:
    	process.communicate()
    	
    	print "\n***Step 3 Done generating the closed itemsets.***\n"
	
    else:
          print "\nNot found Error: Please make sure the fpgrowth directory is at the same directory level as this script (in the src directory).\nAnd the flg2set and fpgrowth scripts are under the fpgrowth/fpgrowth/exec/ directory.\n"
          sys.exit(1)
    
else:
    try:     
        filename1=sys.argv[2]
        filename3=sys.argv[4]
        
        # update the path to the location of the input
        path_to_location=os.path.dirname(filename1)
        
        data=pd.read_csv(filename1,sep=",", header=None)
        data_cols=pd.read_csv(filename3,sep=",", header=None)
        data=data.values
        headers=[x for x in range(len(data[1,:]))]
        headers=np.asarray(headers)
        headers=headers.reshape(1,len(data[1,:]))
        data_=np.concatenate((headers,data))
#         column_names=list(new_data.columns.values)
        np.savetxt(os.path.dirname(filename1)+"/input",data_,delimiter=" ",fmt="%d")
        np.savetxt(os.path.dirname(filename1)+"/column_names",data_cols.values,delimiter="\n",fmt="%s")
        print '\n***Step 1 of Prepocessing successful: The three files "input, column_names, row_names" were written to the root directory. \n'
    except TypeError:
        print "\nType Error: Please provide an input binary file without column names and row names. \n"
        sys.exit(1)

    if os.path.exists("fpgrowth/fpgrowth/ex/flg2set")  and os.path.exists("fpgrowth/fpgrowth/ex/fpgrowth") :
    	# Now we try and call the scripts in the src/fp-growth/fp-growth/exec folder in order to mine the closed itemsets
    	# we call first the flg2set script to convert the input of the user to the right format for the fp-growth algorithm
    	print "Now doing some preprocessing before running the fp-growth algorithm to mine all closed itemsets.\n"
    	
    	print "Converting user input to right format for the fp-growth algorithm.\n"
    	
    	print ("Executing: ./fpgrowth/fpgrowth/ex/flg2set " + path_to_location+"/input"+ " " + path_to_location+"/x1.tab")
    	print ""
    	command="./fpgrowth/fpgrowth/ex/flg2set " + path_to_location+"/input"+ " " + path_to_location+"/x1.tab"
    	process=Popen(command, stdout=PIPE, stderr=None, shell=True)
    	#Launch the shell command:
    	process.communicate()
    	
    	print "\n***Step 2 Done converting to the format required by fp-growth.***\n"
    	
    	# we call first the fpgrowth script to generate the closed itemsets using the fp-growth algorithm
    	print "Now mining the closed itemsets using the fp-growth algorithm.\n"
    	
    	print ("Executing: ./fpgrowth/fpgrowth/ex/fpgrowth -tcm2s"+str(sys.argv[6])+"""k"," -v" """""+"\" " + path_to_location+"/x1.tab"+ " " + path_to_location+"/closed_items")
    	print ""
    	FNULL = open(os.devnull, 'w') 
    	command="./fpgrowth/fpgrowth/ex/fpgrowth -tcm2s"+str(sys.argv[6])+"""k"," -v" """""+"\" " + path_to_location+"/x1.tab"+ " " + path_to_location+"/closed_items"
    	process=Popen(command, stdout=PIPE, stderr=None, shell=True)
    	#Launch the shell command:
    	process.communicate()
    	
    	print "\n***Step 3 Done generating the closed itemsets.***\n"
	
    else:
          print "\nNot found Error: Please make sure the fpgrowth directory is at the same directory level as this script (in the src directory).\nAnd the flg2set and fpgrowth scripts are under the fpgrowth/fpgrowth/exec/ directory. \n"
          sys.exit(1)
    
# once we have the input files and the closed itemsets we can start the exact factorization and save the files
# after this stage, the user can start the interactive part
# get the files
input_f=path_to_location+"/input"
clitems_file=path_to_location+"/closed_items"

# create a folder where we will dump the cPickled files
cPickled_folder_path=path_to_location+"/cPickled"

if not os.path.isdir(cPickled_folder_path):
    os.makedirs(cPickled_folder_path)
    

# start exact factorization
print "Exact Factorization in process ..."
print "This may take several minutes ..."

# we want to keep track of how long this took
start_time = datetime.now() 

# read input data and save some statistics and get the set of items in each transaction
input_mat,trans_sets, indx, nb_cols, nb_rows=ebf.input_read(input_f)

# get the closed itemsets 
cl_items=ebf.closeditems_read(clitems_file,trans_sets, nb_cols)  
  
# write to disk for future use
bkp_closed_f=os.path.dirname(input_f)+'/cPickled/cPickled_closed_sets'
with open(bkp_closed_f, "w") as fp:
    cPickle.dump(cl_items, fp)  

# get the transactions in which each closed itemset is present
trans_items=ebf.trans_items_sets(cl_items,trans_sets)

# write to disk for future use
bkp_trans_f=os.path.dirname(input_f)+'/cPickled/cPickled_trans_sets'
with open(bkp_trans_f, "w") as fp:
    cPickle.dump(trans_items, fp) 
    

# get the number of ones covered by each closed itemset
number_1s=ebf.cover_number(cl_items,trans_items)

# write to disk for future use
bkp_nb1s_f=os.path.dirname(input_f)+'/cPickled/cPickled_number_1s'
with open(bkp_nb1s_f, "w") as fp:
    cPickle.dump(number_1s, fp) 

# get the indices of the 1s that each closed itemset covers
cover=ebf.indices_cover(cl_items,trans_items)

# write to disk for future use
bkp_cover_f=os.path.dirname(input_f)+'/cPickled/cPickled_cover'
with open(bkp_cover_f, "w") as fp:
    cPickle.dump(cover, fp) 

# get the factors after the exact factorization
factor_items=ebf.factors(trans_items,cl_items,number_1s,cover,indx)

# get the reduced index which shows the indices of 1s in the data and the factors that cover them
reduced_indx_cover=ebf.reduced_cover(cover,factor_items)

# get the numbers of indices or 1s that each factor covers alone as the score of that factor
scores=ebf.score(reduced_indx_cover, factor_items)
 
# delete already some factors which cover less than 0.07% of 1s alone
threshold=int((len(indx)*0.07)/100)
 
# By deleting the factors (see below) we will make the following errors
error_from_start=sum([x for x in scores.values() if x<threshold])

# write to disk for future use
bkp_error_f=os.path.dirname(input_f)+'/cPickled/cPickled_error'
with open(bkp_error_f, "w") as fp:
    cPickle.dump(error_from_start, fp) 
    
# deleting factors that cover a number of ones alone less than a threshold  0.07% ..."
# update the scores and factors by deleting the factors which cover less than the threshold number of 1s alone 
scores={k:v for k,v in scores.iteritems() if scores[k]>threshold}
factor_items=[x for x in scores.keys()]

# write to disk for future use
bkp_factors_f=os.path.dirname(input_f)+'/cPickled/cPickled_factor_items'
with open(bkp_factors_f, "w") as fp:
    cPickle.dump(factor_items, fp) 
    
# updating the scores after the deletion of the factors which cover a number ones less than the threshold ..."
# update the scores again by regenerating the scores to reflect the fact that some factors were delted
reduced_indx_cover=ebf.reduced_cover(cover,factor_items)

# write to disk for future use
bkp_reduced_cover_f=os.path.dirname(input_f)+'/cPickled/cPickled_reduced_indx_cover'
with open(bkp_reduced_cover_f, "w") as fp:
    cPickle.dump(reduced_indx_cover, fp) 

scores=ebf.score(reduced_indx_cover, factor_items)

# write to disk for future use
bkp_scores_f=os.path.dirname(input_f)+'/cPickled/cPickled_scores'
with open(bkp_scores_f, "w") as fp:
    cPickle.dump(scores, fp) 
    
print ""
print "***Step 4 Exact Factorization part DONE.***"
print ""
end_time = datetime.now()
print "Exact Factorization took:"
print end_time-start_time
print ""
   
print "The number of factors is:"
print len(factor_items)

    
# filename=sys.argv[-3]
# new_data=pd.io.parsers.read_csv(filename,sep=",",index_col=0)
# data=new_data.values
# headers=[x for x in range(len(data[1,:]))]
# headers=np.asarray(headers)
# headers=headers.reshape(1,len(data[1,:]))
# data_=np.concatenate((headers,data))
# column_names=list(new_data.columns.values)
# row_names=list(new_data.index.values)
# check_rows_names=[x for x in row_names if x not in [0,1]]
# if len(check_rows_names)==0:
#     print "Please select 0 as option [-r 0]. It seems there is no first column in your data with row names"
# else:    
#     np.savetxt("input",data_,delimiter=" ",fmt="%d")
#     np.savetxt("column_names",column_names,delimiter=",",fmt="%s")
#     np.savetxt("row_names",row_names,delimiter=",",fmt="%s")
